//--   File : bwsutil.c (exported version 1.00 on Dec 15, 1993)
//--
//--   Owner: Brad Stumpf
//--

/*#include "win_filt.h"*/
#include <windows.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <direct.h>
#include "exports.h"
#include "cwstart.h"
#include "bwsutil.h"


LIBMAIN


//----------------------------------------------
//--
//--  Export all exportable utility routines.
//--
//----------------------------------------------
void DLL 
_init( void )
{
   //--  LibExports for 'bwsutil.c'
   LibExport( "BWS_notify_beep" );
   LibExport( "BWS_toggle_notify_beep" );
   LibExport( "BOOL BWS_WriteVisibleLines( UINT, LPSTR )" );
   LibExport( "BWS_color_diff_buffer( BOOL )" );
   LibExport( "BWS_BufQNext( BOOL )" );
   LibExport( "BWS_BufQPrev( BOOL )" );
   LibExport( "BWS_EditNextBuffer" );
   LibExport( "BWS_EditPrevBuffer" );

   LibExport( "BWS_BR_edit_file" );
   LibExport( "BWS_BR_read_file" );
   LibExport( "BWS_BR_load_macro" );
   LibExport( "BWS_BR_delete_macro" );
   LibExport( "BWS_output_file" );
   LibExport( "BWS_write_buffer" );
   LibExport( "BWS_BR_qkey" );
}

//----------------------------------------------
//--
//--  Format a common (nice) looking prompt string for file operations.
//--
//----------------------------------------------
LPSTR DLL
_BWS_file_oper_prompt( LPSTR oper )
{
   char cwd[ _MAX_PATH ];
   LPSTR   prompt;

   getcwd( cwd, _MAX_PATH );

   prompt = StrNNew( "", strlen( cwd ) + lstrlen( oper ) + 8 );
   wsprintf( prompt, "[%s]  %s>", oper, cwd );

   return prompt;
}

//----------------------------------------------
//--  
//--  Test system user-flag bit and beep if enabled.
//--
//----------------------------------------------
void DLL
BWS_notify_beep( void )
{
   if ( BWS_TEST_NOTIFY_BEEP )
   {
      SysBeep();
   }
}

//----------------------------------------------
//--  
//--  Toggle system user-flag bit used to support selective beeping.
//--
//----------------------------------------------
void DLL
BWS_toggle_notify_beep( void )
{
   if ( BWS_TEST_NOTIFY_BEEP )
   {
      BWS_RESET_NOTIFY_BEEP;
      MsgMessage( "Notify beep turned OFF." );
   }
   else
   {
      BWS_SET_NOTIFY_BEEP;
      MsgMessage( "Notify beep turned ON." );
   }
}

//----------------------------------------------
//--
//--  Query what the true displayed last line is regardless
//--  the current editing mode (i.e. ASCII, HEX, or COMPACT).
//--
//----------------------------------------------
long DLL
BWS_WinQLastLine( void )
{
   long last;

   if ( !(BufQCurrentBuffer() ) )
      return 0L;

   MarkSavePos();
   MovEndWin();
   last = BufQCurrentLine();
   MarkRestorePos();
   return last;
}

//----------------------------------------------
//--
//--  Re-colorize a (saved) DIFF buffer using the configured
//--  color settings and diff characters.
//--
//----------------------------------------------
void DLL
BWS_color_diff_buffer( BOOL retain_color )
{
   int   add_color = ColorDiffAdd( -1 );
   int   del_color = ColorDiffDel( -1 );
   long  startoff, endoff;
   BOOL  firsttime = TRUE;
   UINT  ch;
   char  strSrch[14];

   if ( !(BufQCurrentBuffer() ) )
      return;

   SysWaitStart();

   MsgMessage( "DIFF ChromaCoding buffer..." );

   BufSetAscii();
   MarkSavePos();
   MovAbs( 1, 1 );

   wsprintf( strSrch, "^[\\%c\\%c]", DIFF_INSERTED, DIFF_DELETED );

   while ( SrchFind( strSrch, SEARCH_REGEX | SEARCH_FORWARD, 0 ) )
   {
      if ( firsttime )
      {
         if ( !retain_color )
         {
            _PosInit( 0L );
            _PosSetColor( ColorText( -1 ), 0x7FFFFFFF );
         }
         firsttime = FALSE;
      }

      if ( SysQBreakPressed() )
      {
         while ( KeyQPressed() )              //causes screen to refresh!!!
                 KeyGetVKey();
         MarkRestorePos();
         SysWaitEnd();
         MsgWarning( "DIFF ChromaCoding aborted with Break key." );
         return;
      }

      _PosInit( -1L );
      startoff = _PosQOffset();

      ch = BufReadChar();
      while ( MovRel( 1, 0 )
           && BufReadChar() == ch );

      if ( BufReadChar() != ch )
         MovRel( -1, 0 );

      _PosInit( -1L );
      _PosNextLine( 0L );
      endoff = _PosQOffset();
      _PosInit( startoff );
      _PosSetColor( ch == DIFF_INSERTED ? add_color : del_color,
                    endoff - startoff );

      MovRel( 1, 0 );
   }

   MarkRestorePos();

   SysWaitEnd();

   if ( !firsttime )
      MsgMessage( "DIFF ChromaCoding complete." );
   else
      MsgPrintf( WARNING_MSG_LEVEL,
                 "No DIFF '%c' or '%c' indicators found in buffer.",
                 DIFF_INSERTED, DIFF_DELETED );
}

//----------------------------------------------
//--
//--  Write only the currently visible lines to a file.
//--     
//--  Param: LineNumCols = number col positions to include (0 to 9)
//--  Param: lpBuf = file to write to.
//--
//--  All params are optional (if no lpBuf param, user will be prompted).
//--
//----------------------------------------------
BOOL DLL
BWS_WriteVisibleLines( UINT LineNumCols, LPSTR lpBuf )
{
#define MAX_LINE_LEN 8190
   LPSTR prompt;
   LPSTR buf    = NIL;
   LPSTR fname  = NIL;
   LPSTR str    = NIL;
   LPSTR ptr;
   int   fileh = -1;
   int   len;
   WORD  ch;
   long  cnt = 0;
   BOOL  bStatus  = FALSE;
   BOOL  bRestPos = FALSE;
   BOOL  bWaiting = FALSE;
   BOOL  bAtEOF;
   char  szLineNum[14];
   LPSTR pLineNum;
   Selection sel;

   if ( !BufQCurrentBuffer()
     || ( BufQSysFlags() & ( BUFFER_HEX_BIN | BUFFER_HEX_ASCII ) ) )
      goto done;

//   if ( ArgCount() < 2 )
//   {
//      lpBuf = NIL;
//      if ( ArgCount() < 1 )
//         LineNumCols = 0;
//   }

   if ( LineNumCols > 9 )
      LineNumCols = 9;

   if ( !(buf = lpBuf) || !*buf )
   {
      prompt = _BWS_file_oper_prompt( "Write Visible" );
      buf    = MsgPromptHistory( "EDITFILE", prompt, "" );

      StrFree( prompt );

      if ( !buf )
         goto done;
   }

   fname = StrNormalizePath( buf );

   if ( !fname || !*fname )
   {
      MsgPrintf( WARNING_MSG_LEVEL, "Invalid path/filename: %s", buf );
      BWS_notify_beep();
      goto done;
   }

   if ( !lpBuf || !*lpBuf )
      StrFree( buf );

   buf = TransformFilename( fname, "%v%p" );

   if ( access( buf, 0 ) )
   {
      if ( SysQFlags() & SYSFLAG_CREATE_DIRS )
      {
         if ( FileCreateDirs( buf ) )
         {
            MsgPrintf( WARNING_MSG_LEVEL,
                       "Unable to create subdir: %s", buf );
            SysBeep();
            goto done;
         }
      }
      else
      {
         MsgPrintf( WARNING_MSG_LEVEL,
                    "Subdir %s does not exist.", buf );
         BWS_notify_beep();
         goto done;
      }
   }

   if ( FileExists( fname ) )
   {
      int   resp;

      prompt = MemAlloc( _fstrlen( fname ) + 32 );

      wsprintf( prompt,  "%s exists.  Overwrite [yn]?", fname );
      BWS_notify_beep();
      resp = MsgConfirm( prompt, "YyNn" );

      MemFree( prompt );

      if ( resp != 'Y' && resp != 'y' )
      {
         MsgMessage( "Visible lines not written." );
         goto done;
      }
   }

   if ( (fileh = FileCreate( fname, OF_WRITE )) == -1 )
   {
      MsgPrintf( WARNING_MSG_LEVEL, "Unable to create file: %s", fname );
      SysBeep();
      goto done;
   }

   SysWaitStart();
   bWaiting = TRUE;

   MsgMessage( "Writing visible lines..." );

   str = (LPSTR)MemAlloc( MAX_LINE_LEN + 2 );

   if ( MarkGetSelInfo( &sel, TRUE ) == SELECTION_NONE )
   {
      sel.begin.line = 1L;
      sel.end.line   = 0x7FFFFFFF;
   }

   MarkSavePos();
   bRestPos = TRUE;

   MovAbs( sel.begin.line, 1L );

   while ( !BufQPastEOF() )
   {
      MovHome();
      _PosInit( -1L );

      ptr = str;
      len = 0;
      ch  = _PosCurrentCharNL();

      if ( LineNumCols > 0 && ch != EOF_CHAR )
      {
         wsprintf( szLineNum, "%9lu:  ", BufQCurrentLine() );
         pLineNum = szLineNum + (long)( 9 - LineNumCols );
         _fstrcpy( ptr, pLineNum );

         len  = ( ch < EOF_CHAR ? 12 : 10 ) - ( 9 - LineNumCols );
         ptr += (long)len;
      }

      while ( len < MAX_LINE_LEN && ch < EOF_CHAR )
      {
        *ptr++ = (char)ch;
        len++;
        ch = _PosNextCharNL();
      }

      bAtEOF = ( ch == EOF_CHAR );

      if ( bAtEOF || len == MAX_LINE_LEN )
         ch = ( BufQUnixEOL() ? EOL1_CHAR : EOL2_CHAR );

      if ( ch == EOL2_CHAR )
      {
         *ptr++ = 0xD;
         *ptr++ = 0xA;
         len += 2;
      }
      else if ( ch == EOL1_CHAR )
      {
         *ptr++ = 0xA;
         len++;
      }

      if ( len > 0 && FileWrite( fileh, (LPVOID) str, len ) == -1 )
      {
         MsgWarning( "Write failed: Disk is possibly full." );
         SysBeep();
         goto done;
      }

      cnt++;

      if ( bAtEOF 
        || BufQCurrentLine() >= sel.end.line
        || !MovDown( 1L ) )
         break;
   }

   if ( BufQCtrlzAppend() )
   {
      *str = 26;
      FileWrite( fileh, (LPVOID) str, 1 );
   }

   MsgPrintf( MESSAGE_MSG_LEVEL, "%lu visible lines written.", cnt );
   bStatus = TRUE;

done:

   if ( fileh != -1 )
      FileClose( fileh );

   if ( bRestPos )
      MarkRestorePos();

   if ( fname )
      StrFree( fname );
   if ( buf )
      StrFree( buf );
   if ( str )
      MemFree( str );

   if ( bWaiting )
      SysWaitEnd();

   if ( bStatus && sel.selType != SELECTION_NONE )
      MarkRemoveSel();

   return bStatus;

#undef MAX_LINE_LEN
}

//----------------------------------------------
//--
//--  These next set of routines implement versions of BufQNext(),
//--  BufQPrev(), EditNextBuffer() and EditPrevBuffer that work
//--  based on the current File List Sorting setting.  This allows
//--  for a more logical (alphabetical) means of cycling thru the
//--  current buffers.
//--
//----------------------------------------------

typedef struct
{
   LPSTR   name;
   HBUFFER hBuf;
} SortBuf;
typedef SortBuf XFAR *LPSortBuf;

//----------------------------------------------
//--
//--  qsort() called routine (used to sort buffer filenames).
//--
//----------------------------------------------
int 
_BWS_sort_bufname_compare( const void *a, const void *b)
{
  LPSortBuf aa = (LPSortBuf)a;
  LPSortBuf bb = (LPSortBuf)b;

  return( _fstrcmp( aa->name, bb->name ) );
}

//----------------------------------------------
//--
//--  Duplicate of Premia's getFileSortName() routine in dlgfunc.c
//--  because it is not exported.
//--
//----------------------------------------------
PRIVATE LPSTR
BWS_getFileSortName(LPSTR name, int mode)
{
   if (mode & FILE_SORT_SHORT_NAMES)
      name = TransformFilename(name, "%r%e\t[ %v%d ]");
   else
   {
      switch(mode & FILE_SORT_MODE_MASK)
      {
      case FILE_SORT_ROOT:
         name = TransformFilename(name, "%r%e   %v%d");
         break;
      case FILE_SORT_EXTENSION:
         name = TransformFilename(name, "%e   %r%v%d");
         break;
      default:
         name = StrNew(name);
         break;
      }
   }
   return(name);
}

//----------------------------------------------
//--
//--  Obtain the next/prev buffer in File List Sort order.
//--
//----------------------------------------------
PRIVATE HBUFFER
BWS_next_buffer( BOOL bPrev, BOOL bSystem )
{
   HBUFFER hCurr;
   HBUFFER hBuf;
   int  bufcnt;
   int  i;
   int  fsmode;
   LPSortBuf lpbaseSB;
   LPSortBuf lpSB;

   if ( (hCurr  = BufQCurrentBuffer()) == 0
     || (bufcnt = BufQBufferCount( BufQIsSystem(hCurr) | bSystem )) < 2
     || (fsmode = FileSortMode(-1) & FILE_SORT_MODE_MASK) == FILE_SORT_NONE
     || !(lpbaseSB = (LPSortBuf)MemAlloc( (int)sizeof(SortBuf) * (++bufcnt))) )
      return( bPrev ? BufQPrev( bSystem ) : BufQNext( bSystem ) );

   hBuf   = hCurr;
   bufcnt = 0;
   lpSB   = lpbaseSB;

   do 
   {
      BufSwitchBuffer( hBuf );
      if ( hBuf == hCurr || bSystem || !BufQIsSystem( hBuf ) )
      {
         if ( (lpSB->name = BufQFilename()) != NIL
           && (lpSB->name = BWS_getFileSortName( lpSB->name, fsmode ) ) != NIL )
         {
            lpSB->hBuf = hBuf;
            lpSB++;
            bufcnt++;
         }
      }
   } while ( (hBuf = BufQNext( TRUE )) != 0 && hBuf != hCurr );

   hBuf = hCurr;
   BufSwitchBuffer( hBuf );

   if ( bufcnt > 1 )
   {
      qsort( (void *)lpbaseSB, bufcnt, sizeof( SortBuf ),
             _BWS_sort_bufname_compare );

      lpSB = lpbaseSB;
      for ( i = 1; i <= bufcnt; i++ )
      {
         if ( lpSB->hBuf == hCurr )
         {
            if ( !bPrev )
               if ( i < bufcnt )
                  lpSB++;
               else
                  lpSB = lpbaseSB;
            else
               if ( i > 1 )
                  lpSB--;
               else
                  lpSB = lpbaseSB + ( bufcnt - 1 );

            break;
         }

         lpSB++;
      }

      if ( i > bufcnt )
         lpSB = lpbaseSB;

      hBuf = lpSB->hBuf;
   }

   lpSB = lpbaseSB;
   for ( i = 1; i <= bufcnt; i++ )
   {
      StrFree( lpSB->name );
      lpSB++;
   }

   MemFree( (void *)lpbaseSB );

   return hBuf;
}

//----------------------------------------------
//--
//--  Mimic Premia's routines but use the File List Sort order.
//--
//----------------------------------------------
HBUFFER DLL
BWS_BufQNext( BOOL bSystem )
{
   return( BWS_next_buffer( FALSE, bSystem ) );
}

HBUFFER DLL
BWS_BufQPrev( BOOL bSystem )
{
   return( BWS_next_buffer( TRUE, bSystem ) );
}

void DLL
BWS_EditNextBuffer( void )
{
   HBUFFER hBuffer;

   if ((hBuffer = BWS_BufQNext(FALSE)) != BufQCurrentBuffer())
   {
      if ((SysQFlags() & SYSFLAG_NEW_WIN_FOR_BUF) ||
            (BufQSysFlags() & BUFFER_SYSTEM))
         SetNewBufferWindow(hBuffer);
      else
         BufSetCurrentBuffer(hBuffer);
      DisplayFileName();
   }
   else
      MsgMessage("No other buffers.");
}

void DLL
BWS_EditPrevBuffer( void )
{
   HBUFFER hBuffer;

   if ((hBuffer = BWS_BufQPrev(FALSE)) != BufQCurrentBuffer())
   {
      if ((SysQFlags() & SYSFLAG_NEW_WIN_FOR_BUF) ||
            (BufQSysFlags() & BUFFER_SYSTEM))
         SetNewBufferWindow(hBuffer);
      else
         BufSetCurrentBuffer(hBuffer);
      DisplayFileName();
   }
   else
      MsgMessage("No other buffers.");
}

//----------------------------------------------
//--
//--  The following are duplicates of some of Premia's BRIEF routines
//--  with support added for _BWS_file_oper_prompt() calling.
//--
//----------------------------------------------

static BOOL
isReadOnly( void )
{
   if (BufQSysFlags() & BUFFER_READ_ONLY)
   {
      MsgMessage( "Cannot modify read-only buffer." );
      SysBeep();
      return TRUE;
   }
   return FALSE;
}

void DLL
BWS_BR_edit_file( void )                             /* Alt-e */
{
// LPSTR   fname = MsgPromptHistory( "EDITFILE", "File: ", "" );
   LPSTR   fname;
   LPSTR   prompt;

   prompt = _BWS_file_oper_prompt( "Edit File" );

   fname = MsgPromptHistory( "EDITFILE", prompt, "" );
   if (fname)
   {
      if (*fname)
      {
         if (BufEditFile( fname ))
            DisplayFileName();
      }
      else
         MsgMessage( "" );
   }      

   StrFree( fname );
   StrFree( prompt );
}

void DLL
BWS_BR_read_file( void )                             /* Alt-r */
{
   LPSTR fname;
   LPSTR prompt;

   CHECK_MIN

   if (isReadOnly())
      return;

   prompt = _BWS_file_oper_prompt( "Read File" );
// fname = MsgPromptHistory( "EDITFILE", "File to read: ", "" );
   fname  = MsgPromptHistory( "EDITFILE", prompt, "" );

   if (fname)
   {
      if (*fname)
      {
         MsgMessage( "Reading..." );
         if (BufInsertFile( fname ))
         {
            MsgWarning( "File not found." );
            BWS_notify_beep();
         }
         else
            MsgMessage( "File read." );
      }
      else
         MsgMessage( "" );
   }      
   StrFree( fname );
   StrFree( prompt );
}

void DLL
BWS_BR_load_macro( void )                            /* F9 */
{
// LPSTR libname = MsgPromptHistory( "MACLOAD", "Library to load: ", NIL );
   LPSTR libname;
   LPSTR prompt;

   prompt = _BWS_file_oper_prompt( "Load DLL" );
   libname = MsgPromptHistory( "MACLOAD", prompt, NIL );

   if (libname && *libname)
   {
      if (LibPreLoad( libname ) > 0)
         MsgMessage( "" );
      else
      {
         MsgMessage( "Unable to load library." );
         SysBeep();
      }
   }
   StrFree( libname );
   StrFree( prompt );
}

void DLL
BWS_BR_delete_macro( void )                          /* Shift-F9 */
{
// LPSTR libname = MsgPrompt( "Library to unload: ", NIL, NIL );
   LPSTR libname;
   LPSTR prompt;

   prompt = _BWS_file_oper_prompt( "UnLoad DLL" );
   libname = MsgPromptHistory( "MACLOAD", prompt, NIL );


   if (libname && *libname)
   {
      if (LibUnLoad( libname ))
         MsgMessage( "" );
      else
      {
         MsgMessage( "Unable to unload library." );
         SysBeep();
      }
   }

   StrFree( libname );
   StrFree( prompt );
}

void DLL
BWS_output_file( void )
{
   LPSTR oname;
   LPSTR prompt;

   if ((oname = BufQOutputName()) == NIL)
      oname = BufQFilename();

   prompt = _BWS_file_oper_prompt( "Change Output File" );
// oname = MsgPrompt( "Enter new output file name: ", oname, NIL);
   oname = MsgPrompt( prompt, oname, NIL);

   if (oname && *oname)
   {
      LPSTR buf;

      buf = StrNormalizePath( oname );

      if ( !buf )
      {
         MsgPrintf( WARNING_MSG_LEVEL, "Invalid path/filename: %s", oname );
         BWS_notify_beep();
      }
      else
      {
         LPSTR dr   = NIL;
         LPSTR path = NIL;
         BOOL  chg_name = TRUE;

         StrSplitpath( buf, &dr, &path, NULL, NULL );
         dr = StrAppend( dr, ":" );
         dr = StrAppend( dr, path );

         if ( access( dr, 0 ) )
         {
            LPSTR cstr = (LPSTR)MemAlloc( _fstrlen( dr ) + 50 );
            int   resp;

            chg_name = FALSE;

            wsprintf( cstr, "Subdir %s does not exist.  Create [yn]?", dr );
            SysBeep();
            resp = MsgConfirm( cstr, "YyNn" );

            MemFree( cstr );

            if ( resp == 'Y' || resp == 'y' )
            {
               if ( FileCreateDirs( dr ) )
               {
                  MsgPrintf( WARNING_MSG_LEVEL,
                             "Unable to create subdir: %s", dr );
                  SysBeep();
               }
               else
               {
                 chg_name = TRUE;
               }
            }
         }

         if ( chg_name )
         {
            HBUFFER currbuf = BufQCurrentBuffer();

            BufSetOutputName( buf );

            BufSwitchBuffer( BufQModelBuffer() );
            BufSetAttributes( currbuf );
            BufSwitchBuffer( currbuf );
            BufSetModified( TRUE );
//          BufSetReadOnly( FALSE );

            if ( FileExists( buf ) )
            {
               MsgPrintf( WARNING_MSG_LEVEL, "WARNING: %s exists!", buf );
               SysBeep();
            }
            else
               DisplayFileName();
         }

         StrFree( dr );
         StrFree( path );
      }

      if ( buf )
         StrFree( buf );
   }

   if ( oname )
      StrFree( oname );
   StrFree( prompt );
}

void DLL
BWS_write_buffer( void )
{
   LPSTR oname;

   if ( MarkQSelType() )
   {
//BWS LPSTR fname = MsgPrompt( "Write block as: ", "", NIL );
      LPSTR prompt = _BWS_file_oper_prompt( "Write Block" );
      LPSTR buf    = MsgPromptHistory( "EDITFILE", prompt, "" );
      LPSTR fname  = NIL;

      StrFree( prompt );

      if ( buf )
      {
         fname = StrNormalizePath( buf );

         if ( !fname )
         {
            MsgPrintf( WARNING_MSG_LEVEL, "Invalid path/filename: %s", buf );
            BWS_notify_beep();
         }

         StrFree( buf );
      }

      if (fname)
      {
         if ( FileExists( fname ) )
         {
            int   resp;

            prompt = MemAlloc( _fstrlen( fname ) + 32 );

            wsprintf( prompt,  "%s exists.  Overwrite [yn]?", fname );
            BWS_notify_beep();
            resp = MsgConfirm( prompt, "YyNn" );

            MemFree( prompt );

            if ( resp != 'Y' && resp != 'y' )
            {
               *fname = '\0';
               MsgMessage( "Block not written." );
            }
         }

         if ( *fname )
            if (BufWriteSelection( fname ))
            {
               MsgPrintf( WARNING_MSG_LEVEL, "Invalid path/filename: %s", fname );
               SysBeep();
            }
            else
            {
               MsgMessage( "Write Block successful." );
               MarkRemoveSel();
            }

         StrFree( fname );
      }
   }
   else
   {
      if (BufQModified())
      {
         if ((oname = BufQOutputName()) == NIL)
            oname = BufQFilename();
         MsgPrintf( MESSAGE_MSG_LEVEL, "Writing %s", oname);
         if (!BufWrite())
            MsgMessage( "Write successful." );
         else
         {
            MsgPrintf( MESSAGE_MSG_LEVEL, "Unable to write file: %s", oname);
            SysBeep();
         }
      }
      else
      {
         MsgWarning( "File has not been modified -- not written." );
         BWS_notify_beep();
      }
   }
}

//----------------------------------------------
//--
//--  Fixed and better version of Premia's BR_qkey in BRIEF.C.
//--
//----------------------------------------------
void DLL
BWS_BR_qkey( void )
{
    int lvlSav = MsgLevel(0);
    int code, prev_code;
    LPSTR pAss, pStr;
    char  szKeyCode[30];

//  MsgMessage("You are now in KEY QUERY mode. Press ESC to resume edit");
    MsgMessage("KEY QUERY mode:  Press <Esc><Esc> to resume..." );
    do
    {
        prev_code = code;
        code = KeyGetVKey();

        if ((pStr = KeyCodeToStr(code)) == NULL)
            pStr = "NULL";

        wsprintf(szKeyCode,"#%d", code);
        if ((pAss = KmapQAssignment(szKeyCode)) == NULL)
            pAss = "Not assigned";

//      MsgPrintf(0, "%d => %s => %s", code, pStr, pAss);
        MsgPrintf(0, "0x%04x => %s => %s", code, pStr, pAss);
    }
    while( code != 0x1B                 // Quit on ESCape
        || code != prev_code );         // Twice...BWS

    MsgMessage( "Editing resumed." );
    MsgLevel(lvlSav);

}

